#include "logreader.h"
#include <sqlextract/util.h>
#include <stdlib.h>
#include <stdio.h>
#include <sqlextract/tables/table.h>
#include <sqlextract/filter/arbval.h>
#include <string.h>
#include <sqlextract/strtools.h>

#include <unicode/utypes.h>
#include <unicode/ucnv_err.h>
#include <unicode/uenum.h>
#include <unicode/ucnv.h>

/* Maximum number of columns in a log line on disk */
#define LOGREADER_MAX_COLUMNS 64

/* Number of columns available to SQL */
#define LOGREADER_NUM_COLUMNS 11

struct logreader_t {
	int *columnmap;
	int columnmap_len;

	FILE *fp;

	char *logline[LOGREADER_MAX_COLUMNS];
	int nfields;

	char *linebuf;
	size_t linebuf_len;

	struct arbval_t *curfield;
};

struct logreader_fielddesc_t {
	const char *name;
	int index;
	void (*interpreter)(struct arbval_t *, const char *);
};

static void *logreader_create(struct strid_t *si, void *fp);
static void logreader_destroy(void *lr);
static void *logreader_create_iterator(void *lr);
static int logreader_fetch_row(void *it);
static struct arbval_t *logreader_fetch_field(void *it, int fieldname);
static void logreader_destroy_iterator(void *it);

static void logreader_interpreter_string(struct arbval_t *result,
		const char *str);
static void logreader_interpreter_integer(struct arbval_t *result,
		const char *str);
static void logreader_interpreter_timestamp(struct arbval_t *result,
		const char *str);

/* Special purpose interpreters */
static void logreader_interpreter_facility(struct arbval_t *result,
		const char *str);
static void logreader_interpreter_severity(struct arbval_t *result,
		const char *str);

struct table_desc_t table_logreader = { logreader_create, logreader_destroy,
		logreader_create_iterator, logreader_fetch_row, logreader_fetch_field,
		logreader_destroy_iterator };

static struct logreader_fielddesc_t logreader_columns[LOGREADER_NUM_COLUMNS + 1] = { { "ip", 0,
		logreader_interpreter_string }, // varchar(110)",
		{ "rtime", 1, logreader_interpreter_timestamp }, // timestamp",
		{ "facility", 2, logreader_interpreter_facility }, // integer",
		{ "severity", 2, logreader_interpreter_severity }, // integer",
		{ "mtime", 3, logreader_interpreter_timestamp }, // timestamp",
		{ "host", 4, logreader_interpreter_string }, // varchar(129)",
		{ "ident", 5, logreader_interpreter_string }, // varchar(80)",
		{ "pid", 6, logreader_interpreter_string }, // varchar(128)",
		{ "event", 7, logreader_interpreter_integer }, // integer",
		{ "msg", 8, logreader_interpreter_string }, // text",
		{ "fts", 9, logreader_interpreter_string }, // text",
		{ NULL, 0, NULL } };

static void *logreader_create(struct strid_t *si, void *fp) {
	struct logreader_t *lr;
	FILE *fpp = fp;
	int max_id;
	int id;
	int i;

	lr = (struct logreader_t *) lf_calloc(1, sizeof(struct logreader_t));
	lr->fp = fpp;

	max_id = 0;
	for (i = 0; logreader_columns[i].name; i++) {
		id = strid_get(si, logreader_columns[i].name);
		if (id > max_id)
			max_id = id;
	}
	lr->columnmap_len = max_id + 1;
	lr->columnmap = lf_calloc(lr->columnmap_len, sizeof(int));
	for (i = 0; i < lr->columnmap_len; i++) {
		lr->columnmap[i] = -1;
	}
	for (i = 0; logreader_columns[i].name; i++) {
		id = strid_get(si, logreader_columns[i].name);
		lr->columnmap[id] = i;
	}

	lr->nfields = 0;
	lr->curfield = arbval_create_null();

	return lr;
}

static void logreader_destroy(void *lr) {
	struct logreader_t *lrp = lr;
	if (lrp == NULL)
		return;
	if (lrp->linebuf)
		free(lrp->linebuf);
	if (lrp->columnmap)
		free(lrp->columnmap);
	if (lrp->curfield)
		arbval_destroy(lrp->curfield);
	free(lrp);
}

static void *logreader_create_iterator(void *lr) {
	struct logreader_t *lrp = lr;
	/* TODO: Make it possible to seek in logs, if reading from file
	 * out of scope if logs only comes from stdin
	 */
	return lrp;
}

static void logreader_destroy_iterator(void *it) {
	/* Since iterator at the moment is the logreader struct, don't do anything */
}

static int logreader_fetch_row(void *it) {
	struct logreader_t *lrp = it;

	size_t nchars;
	char *tokptr;
	size_t nfields;

	nchars = getline(&(lrp->linebuf), &(lrp->linebuf_len), lrp->fp);
	if (nchars == -1) {
		/* Error, probably end of file */
		return 0;
	}

	/* Remove trailing linebreaks */
	while (nchars > 0 && lrp->linebuf[nchars - 1] == '\n') {
		nchars--;
	}
	lrp->linebuf[nchars] = '\0';

	/* Split up lines */
	lrp->logline[0] = lrp->linebuf;

	nfields = 1;
	for (tokptr = lrp->linebuf; *tokptr != '\0'; tokptr++) {
		if (*tokptr == 0x1e) {
			*tokptr = '\0';
			lrp->logline[nfields++] = tokptr + 1;
		}
	}
	lrp->nfields = nfields;

	return 1;
}

static struct arbval_t *logreader_fetch_field(void *it, int fieldname) {
	struct logreader_t *lrp = it;
	int col_id;

	if (fieldname >= lrp->columnmap_len) {
		arbval_typecast_null(lrp->curfield);
		return lrp->curfield;
	}

	col_id = lrp->columnmap[fieldname];
	if (col_id < 0) {
		arbval_typecast_null(lrp->curfield);
		return lrp->curfield;
	}

	if (col_id >= LOGREADER_NUM_COLUMNS) {
		arbval_typecast_null(lrp->curfield);
		return lrp->curfield;
	}

	if (logreader_columns[col_id].index >= lrp->nfields) {
		arbval_typecast_null(lrp->curfield);
		return lrp->curfield;
	}

	(*logreader_columns[col_id].interpreter)(lrp->curfield,
			lrp->logline[logreader_columns[col_id].index]);

	return lrp->curfield;
}

static void logreader_interpreter_string(struct arbval_t *result,
		const char *str) {
	long len = strlen(str)*2+1;

	arbval_typecast_null(result);
	result->type = ARBVAL_TYPE_STRING;
	result->value.string = lf_calloc(1, len);

	/*
	 * We need to assure that it's a valid UTF-8 string, so we have a
	 * determinstic charset within the program
	 */
	{
		const char *from_charsets[] = {"UTF-8", "ISO8859-1", NULL};
		int i;
		for(i=0;from_charsets[i];i++) {
			UErrorCode err = U_ZERO_ERROR;
			ucnv_convert("UTF-8", from_charsets[i], result->value.string, len, str, -1, &err);
			if(U_SUCCESS(err)) {
				/* We managed to convert, exit */
				return;
			}
		}
		/* No converters finished, thus, we don't know anything about the data. Return NULL */
		arbval_typecast_null(result);
	}
}
static void logreader_interpreter_integer(struct arbval_t *result,
		const char *str) {
	logreader_interpreter_string(result, str);
	arbval_typecast_integer(result);
}
static void logreader_interpreter_timestamp(struct arbval_t *result,
		const char *str) {

	long ts = parsestrtime(str);

	if (ts >= 0) {
		arbval_typecast_null(result);
		result->type = ARBVAL_TYPE_INTEGER;
		result->value.integer = ts;
	} else {
		logreader_interpreter_string(result, str);
	}
}

static void logreader_interpreter_facility(struct arbval_t *result,
		const char *str) {
	logreader_interpreter_string(result, str);
	arbval_typecast_integer(result);
	result->value.integer >>= 3;
}
static void logreader_interpreter_severity(struct arbval_t *result,
		const char *str) {
	logreader_interpreter_string(result, str);
	arbval_typecast_integer(result);
	result->value.integer &= 7;
}
